package install_test

import (
	"bytes"
	"path/filepath"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/ginkgo/extensions/table"
	. "github.com/onsi/gomega"

	"github.com/kumahq/kuma/app/kumactl/cmd"
	kumactl_cmd "github.com/kumahq/kuma/app/kumactl/pkg/cmd"
	"github.com/kumahq/kuma/pkg/util/test"
	kuma_version "github.com/kumahq/kuma/pkg/version"
)

var _ = Describe("kumactl install control-plane", func() {

	var backupBuildInfo kuma_version.BuildInfo
	BeforeEach(func() {
		backupBuildInfo = kuma_version.Build
	})
	AfterEach(func() {
		kuma_version.Build = backupBuildInfo
	})

	BeforeEach(func() {
		kuma_version.Build = kuma_version.BuildInfo{
			Version:   "0.0.1",
			GitTag:    "v0.0.1",
			GitCommit: "91ce236824a9d875601679aa80c63783fb0e8725",
			BuildDate: "2019-08-07T11:26:06Z",
		}
	})

	var stdout *bytes.Buffer
	var stderr *bytes.Buffer

	BeforeEach(func() {
		stdout = &bytes.Buffer{}
		stderr = &bytes.Buffer{}
	})

	type testCase struct {
		extraArgs  []string
		goldenFile string
	}

	DescribeTable("should generate Kubernetes resources",
		func(given testCase) {
			// setup - override some values to have stable YAMLs regardless of changes in versions and autogenerated values
			rootCtx := kumactl_cmd.DefaultRootContext()
			rootCtx.Runtime.NewAPIServerClient = test.GetMockNewAPIServerClient()
			rootCtx.InstallCpContext.Args.ControlPlane_image_tag = "0.0.1"
			rootCtx.InstallCpContext.Args.DataPlane_image_tag = "0.0.1"
			rootCtx.InstallCpContext.Args.DataPlane_initImage_tag = "0.0.1"

			// given
			rootCmd := cmd.NewRootCmd(rootCtx)
			rootCmd.SetArgs(append(
				[]string{
					"install",
					"control-plane",
					"--tls-general-secret", "general-tls-secret",
					"--tls-general-ca-bundle", "XYZ",
				},
				given.extraArgs...,
			))
			rootCmd.SetOut(stdout)
			rootCmd.SetErr(stderr)

			// when
			err := rootCmd.Execute()

			// then command succeed
			Expect(err).ToNot(HaveOccurred())
			Expect(stderr.String()).To(BeEmpty())

			// and output matches golden files
			actual := stdout.Bytes()
			ExpectMatchesGoldenFiles(actual, filepath.Join("testdata", given.goldenFile))
		},
		Entry("should generate Kubernetes resources with default settings", testCase{
			extraArgs: []string{
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.defaults.golden.yaml",
		}),
		Entry("should override default env-vars with values supplied", testCase{
			extraArgs: []string{
				"--env-var", "KUMA_DEFAULTS_SKIP_MESH_CREATION=true",
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.override-env-vars.golden.yaml",
		}),
		Entry("should generate Kubernetes resources with custom settings", testCase{
			extraArgs: []string{
				"--namespace", "kuma",
				"--image-pull-policy", "Never",
				"--control-plane-version", "greatest",
				"--control-plane-registry", "kuma-ci",
				"--control-plane-service-name", "kuma-ctrl-plane",
				"--injector-failure-policy", "Crash",
				"--dataplane-registry", "kuma-ci",
				"--dataplane-version", "greatest",
				"--dataplane-init-registry", "kuma-ci",
				"--dataplane-init-version", "greatest",
				"--tls-api-server-secret", "api-server-secret",
				"--tls-api-server-client-certs-secret", "api-server-client-secret",
				"--tls-kds-global-server-secret", "kds-global-secret",
				"--tls-kds-zone-client-secret", "kds-ca-secret",
				"--tls-general-ca-secret", "general-tls-secret-ca",
				"--mode", "zone",
				"--kds-global-address", "grpcs://192.168.0.1:5685",
				"--zone", "zone-1",
				"--use-node-port",
				"--without-kubernetes-connection",
				"--experimental-meshgateway",
			},
			goldenFile: "install-control-plane.overrides.golden.yaml",
		}),
		Entry("should generate Kubernetes resources with CNI plugin", testCase{
			extraArgs: []string{
				"--cni-enabled",
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.cni-enabled.golden.yaml",
		}),
		Entry("should generate Kubernetes resources for Global", testCase{
			extraArgs: []string{
				"--mode", "global",
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.global.golden.yaml",
		}),
		Entry("should generate Kubernetes resources for Zone", testCase{
			extraArgs: []string{
				"--mode", "zone",
				"--zone", "zone-1",
				"--kds-global-address", "grpcs://192.168.0.1:5685",
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.zone.golden.yaml",
		}),
		Entry("should generate Kubernetes resources with Ingress enabled", testCase{
			extraArgs: []string{
				"--ingress-enabled",
				"--ingress-drain-time", "60s",
				"--ingress-use-node-port",
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.with-ingress.golden.yaml",
		}),
		Entry("should generate Kubernetes resources with Egress enabled", testCase{
			extraArgs: []string{
				"--egress-enabled",
				"--egress-drain-time", "60s",
				"--without-kubernetes-connection",
			},
			goldenFile: "install-control-plane.with-egress.golden.yaml",
		}),
	)

	type errTestCase struct {
		extraArgs []string
		errorMsg  string
	}
	DescribeTable("should fail to install control plane",
		func(given errTestCase) {
			// given
			rootCmd := test.DefaultTestingRootCmd()
			rootCmd.SetArgs(append([]string{"install", "control-plane"}, given.extraArgs...))
			rootCmd.SetOut(stdout)
			rootCmd.SetErr(stderr)

			// when
			err := rootCmd.Execute()

			// then
			Expect(err).To(HaveOccurred())
			Expect(err.Error()).To(Equal(given.errorMsg))
		},
		Entry("--mode is unknown", errTestCase{
			extraArgs: []string{"--mode", "test"},
			errorMsg:  "invalid mode. Available modes: standalone, zone, global",
		}),
		Entry("--kds-global-address is missing when installing zone", errTestCase{
			extraArgs: []string{"--mode", "zone", "--zone", "zone-1"},
			errorMsg:  "--kds-global-address is mandatory with `zone` mode",
		}),
		Entry("--kds-global-address is not valid URL", errTestCase{
			extraArgs: []string{"--kds-global-address", "192.168.0.1:1234", "--mode", "zone", "--zone", "zone-1"},
			errorMsg:  "--kds-global-address is not valid URL. The allowed format is grpcs://hostname:port",
		}),
		Entry("--kds-global-address has no grpcs scheme", errTestCase{
			extraArgs: []string{"--kds-global-address", "http://192.168.0.1:1234", "--mode", "zone", "--zone", "zone-1"},
			errorMsg:  "--kds-global-address should start with grpcs://",
		}),
		Entry("--kds-global-address is used with standalone", errTestCase{
			extraArgs: []string{"--kds-global-address", "192.168.0.1:1234", "--mode", "standalone"},
			errorMsg:  "--kds-global-address can only be used when --mode=zone",
		}),
		Entry("--tls-general-secret without --tls-general-ca-bundle", errTestCase{
			extraArgs: []string{"--tls-general-secret", "sec"},
			errorMsg:  "--tls-general-secret and --tls-general-ca-bundle must be provided at the same time",
		}),
	)
})
